@@ -3,8 +3,7 @@
 import os
 from .model_interfaces import PhygnnModel, RandomForestModel, TfModel
 from .phygnn import PhysicsGuidedNeuralNetwork, p_fun_dummy
-from .utilities import Layers, HiddenLayers
-from .utilities import tf_isin, tf_log10
+from .utilities import Layers, HiddenLayers, PreProcess, tf_isin, tf_log10
 from phygnn.version import __version__
 
 PHYGNNDIR = os.path.dirname(os.path.realpath(__file__))
