@@ -6,8 +6,10 @@ from abc import ABC
 import logging
 import numpy as np
 import pandas as pd
+import tensorflow as tf
 from warnings import warn
 
+from phygnn.utilities import TF2
 from phygnn.utilities.pre_processing import PreProcess
 
 logger = logging.getLogger(__name__)
@@ -17,14 +19,14 @@ class ModelBase(ABC):
     """
     Base Model Interface
     """
-
     def __init__(self, model, feature_names=None, label_names=None,
-                 norm_params=None, normalize=(True, False)):
+                 norm_params=None, normalize=(True, False),
+                 one_hot_categories=None):
         """
         Parameters
         ----------
         model : OBJ
-            Sci-kit learn or tensorflow model
+            Initialized model object
         feature_names : list
             Ordered list of feature names.
         label_names : list
@@ -39,19 +41,22 @@ class ModelBase(ABC):
             - False means don't normalize either
             - Tuple of flags (normalize_feature, normalize_label)
             by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
         """
         self._model = model
 
         if isinstance(feature_names, str):
             feature_names = [feature_names]
-        elif isinstance(feature_names, np.ndarray):
+        elif isinstance(feature_names, (np.ndarray, pd.Index)):
             feature_names = feature_names.tolist()
 
         self._feature_names = feature_names
 
         if isinstance(label_names, str):
             label_names = [label_names]
-        elif isinstance(label_names, np.ndarray):
+        elif isinstance(label_names, (np.ndarray, pd.Index)):
             label_names = label_names.tolist()
 
         self._label_names = label_names
@@ -60,6 +65,10 @@ class ModelBase(ABC):
 
         self._norm_params = norm_params
         self._normalize = self._parse_normalize(normalize)
+        if one_hot_categories is not None:
+            PreProcess.check_one_hot_categories(one_hot_categories)
+
+        self._one_hot_categories = one_hot_categories
 
     def __repr__(self):
         msg = "{}:\n{}".format(self.__class__.__name__, self.model_summary)
@@ -288,6 +297,60 @@ class ModelBase(ABC):
 
         return stdevs
 
+    @property
+    def input_feature_names(self):
+        """
+        Input feature names
+
+        Return
+        ------
+        list
+        """
+        if self._one_hot_categories is None:
+            input_feature_names = self.feature_names
+        else:
+            input_feature_names = list(set(self.feature_names)
+                                       - set(self.one_hot_feature_names)
+                                       | set(self.one_hot_input_feature_names))
+
+        return input_feature_names
+
+    @property
+    def one_hot_input_feature_names(self):
+        """
+        Input feature names to be one-hot encoded
+
+        Return
+        ------
+        list
+        """
+        return list(self.one_hot_categories.keys())
+
+    @property
+    def one_hot_feature_names(self):
+        """
+        One-hot encoded feature names
+
+        Return
+        ------
+        list
+        """
+        return [i for sub in self.one_hot_categories.values() for i in sub]
+
+    @property
+    def one_hot_categories(self):
+        """
+        categories to use for one-hot encoding
+
+        Returns
+        -------
+        dict
+        """
+        if self._one_hot_categories is None:
+            return {}
+        else:
+            return self._one_hot_categories
+
     @staticmethod
     def _parse_normalize(normalize):
         """
@@ -357,6 +420,21 @@ class ModelBase(ABC):
 
         return out
 
+    @staticmethod
+    def seed(s=0):
+        """
+        Set the random seed for reproducible results.
+        Parameters
+        ----------
+        s : int
+            Random number generator seed
+        """
+        np.random.seed(s)
+        if TF2:
+            tf.random.set_seed(s)
+        else:
+            tf.random.set_random_seed(s)
+
     @staticmethod
     def _parse_data(data, names=None):
         """
@@ -377,7 +455,7 @@ class ModelBase(ABC):
             List of data item names
         """
         if isinstance(data, pd.DataFrame):
-            names = data.columns.values.tolist()
+            names = data.columns.tolist()
             data = data.values
         elif isinstance(data, dict):
             names = list(data.keys())
@@ -413,6 +491,34 @@ class ModelBase(ABC):
 
         return n
 
+    @staticmethod
+    def make_one_hot_feature_names(feature_names, one_hot_categories):
+        """
+        Update feature_names after one-hot encoding
+
+        Parameters
+        ----------
+        feature_names : list
+            Input feature names
+        one_hot_categories : dict
+            Features to one-hot encode using given categories
+
+        Returns
+        -------
+        one_hot_feature_names : list
+            Updated list of feature names with one_hot categories
+        """
+        one_hot_feature_names = feature_names.copy()
+        for name, categories in one_hot_categories.items():
+            if name in one_hot_feature_names:
+                one_hot_feature_names.remove(name)
+
+            for c in categories:
+                if c not in one_hot_feature_names:
+                    one_hot_feature_names.append(c)
+
+        return one_hot_feature_names
+
     def get_norm_params(self, names):
         """
         Get means and stdevs for given feature/label names
@@ -435,10 +541,10 @@ class ModelBase(ABC):
             means.append(self.get_mean(name))
             stdevs.append(self.get_stdev(name))
 
-        if not all(set(means)):
+        if None in means:
             means = None
 
-        if not all(set(stdevs)):
+        if None in stdevs:
             stdevs = None
 
         return means, stdevs
@@ -499,50 +605,26 @@ class ModelBase(ABC):
         """
         norm_items = {}
         for key, value in items.items():
-            mean = self.get_mean(key)
-            stdev = self.get_stdev(key)
-            update = mean is None or stdev is None
-            try:
-                value, mean, stdev = PreProcess.normalize(value, mean=mean,
-                                                          stdev=stdev)
-                if update:
-                    norm_params = {key: {'mean': mean, 'stdev': stdev}}
-                    self._norm_params.update(norm_params)
-            except Exception as ex:
-                msg = "Could not normalize {}:\n{}".format(key, ex)
-                logger.warning(msg)
-                warn(msg)
+            if key not in self.one_hot_feature_names:
+                mean = self.get_mean(key)
+                stdev = self.get_stdev(key)
+                update = mean is None or stdev is None
+                try:
+                    value, mean, stdev = PreProcess.normalize(value,
+                                                              mean=mean,
+                                                              stdev=stdev)
+                    if update:
+                        norm_params = {key: {'mean': mean, 'stdev': stdev}}
+                        self._norm_params.update(norm_params)
+                except Exception as ex:
+                    msg = "Could not normalize {}:\n{}".format(key, ex)
+                    logger.warning(msg)
+                    warn(msg)
 
             norm_items[key] = value
 
         return norm_items
 
-    def _normalize_df(self, df):
-        """
-        Normalize DataFrame
-
-        Parameters
-        ----------
-        df : pandas.DataFrame
-            DataFrame of features/label to normalize
-
-        Returns
-        -------
-        norm_df : pandas.DataFrame
-            Normalized features/label
-        """
-        means, stdevs = self.get_norm_params(df.columns)
-        update = means is None or stdevs is None
-
-        norm_df, means, stdevs = PreProcess.normalize(df, mean=means,
-                                                      stdev=stdevs)
-        if update:
-            for i, c in enumerate(df.columns):
-                norm_params = {c: {'mean': means[i], 'stdev': stdevs[i]}}
-                self._norm_params.update(norm_params)
-
-        return norm_df
-
     def _normalize_arr(self, arr, names):
         """
         Normalize array and save normalization parameters to given names
@@ -598,7 +680,12 @@ class ModelBase(ABC):
         if isinstance(data, dict):
             data = self._normalize_dict(data)
         elif isinstance(data, pd.DataFrame):
-            data = self._normalize_df(data)
+            if self.one_hot_feature_names:
+                cols = [c for c in data if c not in self.one_hot_feature_names]
+                data.loc[:, cols] = self._normalize_arr(
+                    data.loc[:, cols].values, cols)
+            else:
+                data.loc[:] = self._normalize_arr(data.values, data.columns)
         elif isinstance(data, (list, np.ndarray)):
             if names is None:
                 msg = ('Names of items must be supplied to nomralize data '
@@ -606,7 +693,14 @@ class ModelBase(ABC):
                 logger.error(msg)
                 raise RuntimeError(msg)
             else:
-                data = self._normalize_arr(data, names)
+                if self.one_hot_feature_names:
+                    idx = [i for i, f in enumerate(names)
+                           if f not in self.one_hot_feature_names]
+                    norm_names = np.array(names)[idx]
+                    data[:, idx] = self._normalize_arr(data[:, idx],
+                                                       norm_names)
+                else:
+                    data = self._normalize_arr(data, names)
         else:
             msg = "Cannot normalize data of type: {}".format(type(data))
             logger.error(msg)
@@ -741,25 +835,28 @@ class ModelBase(ABC):
 
         return data
 
-    def _check_one_hot_norm_params(self, one_hot_features):
+    def _check_one_hot_feature_names(self, feature_names):
         """
-        Check one hot feature normalization parameters to ensure they are
-        {mean: 0, stdev: 1} to prevent normalization
+        Check one_hot_feature_names, update feature_names to remove features
+        that were one-hot encoded and add in new one-hot features if needed
 
         Parameters
         ----------
-        one_hot_features : list
-            list of one hot features
+        feature_names : list
+            Input feature names
         """
-        for feature in one_hot_features:
-            mean = self.get_mean(feature)
-            stdev = self.get_stdev(feature)
-            if mean != 0 and stdev != 1:
-                norm_params = {feature: {'mean': 0, 'stdev': 1}}
-                self._norm_params.update(norm_params)
+        one_hot_feature_names = self.make_one_hot_feature_names(
+            feature_names, self.one_hot_categories)
+        if one_hot_feature_names != self.feature_names:
+            check_names = feature_names.copy()
+            if self.label_names is not None:
+                check_names += self.label_names
 
-    def _parse_features(self, features, names=None, process_one_hot=True,
-                        **kwargs):
+            PreProcess.check_one_hot_categories(self.one_hot_categories,
+                                                feature_names=check_names)
+            self._feature_names = one_hot_feature_names
+
+    def _parse_features(self, features, names=None, **kwargs):
         """
         Parse features
 
@@ -769,8 +866,6 @@ class ModelBase(ABC):
             Features to train on or predict from
         names : list, optional
             List of feature names, by default None
-        process_one_hot : bool, optional
-            Check for and process one-hot variables, by default True
         kwargs : dict, optional
             kwargs for PreProcess.one_hot
 
@@ -788,30 +883,30 @@ class ModelBase(ABC):
             logger.error(msg)
             raise RuntimeError(msg)
 
-        if self.feature_names is not None:
-            if features.shape[1] != len(self.feature_names):
-                msg = ('data has {} features but expected {}'
-                       .format(features.shape[1], self.feature_dims))
-                logger.error(msg)
-                raise RuntimeError(msg)
-
-        if self._feature_names is None:
+        if self.feature_names is None:
             self._feature_names = feature_names
+
+        check = (self.one_hot_categories is not None
+                 and all(np.isin(feature_names, self.input_feature_names)))
+        if check:
+            self._check_one_hot_feature_names(feature_names)
+            kwargs.update({'feature_names': feature_names,
+                           'categories': self.one_hot_categories})
+            features = PreProcess.one_hot(features, **kwargs)
         elif self.feature_names != feature_names:
             msg = ('Expecting features with names: {}, but was provided with: '
-                   '{}!'.format(feature_names, self.feature_names))
+                   '{}!'.format(self.feature_names, feature_names))
             logger.error(msg)
             raise RuntimeError(msg)
 
-        if process_one_hot:
-            kwargs.update({'return_ind': True})
-            features, one_hot_ind = PreProcess.one_hot(features, **kwargs)
-            if one_hot_ind:
-                one_hot_features = [self.feature_names[i] for i in one_hot_ind]
-                self._check_one_hot_norm_params(one_hot_features)
-
         if self.normalize_features:
-            features = self.normalize(features, names=feature_names)
+            features = self.normalize(features, names=self.feature_names)
+
+        if features.shape[1] != self.feature_dims:
+            msg = ('data has {} features but expected {}'
+                   .format(features.shape[1], self.feature_dims))
+            logger.error(msg)
+            raise RuntimeError(msg)
 
         return features
 
@@ -879,7 +974,24 @@ class ModelBase(ABC):
             parse_kwargs = {}
 
         if isinstance(features, np.ndarray):
-            parse_kwargs.update({"names": self.feature_names})
+            n_features = features.shape[1]
+            if n_features == self.feature_dims:
+                kwargs = {"names": self.feature_names}
+                logger.debug('Parsing features with feature_names: {}'
+                             .format(self.feature_names))
+            elif n_features == len(self.input_feature_names):
+                kwargs = {"names": self.input_feature_names}
+                logger.debug('Parsing features with input_feature_names: {}'
+                             .format(self.input_feature_names))
+            else:
+                msg = ('Number of features provided ({}) does not match number'
+                       ' of model features ({}) or number of input features '
+                       '({})'.format(n_features, self.feature_dims,
+                                     len(self.input_feature_names)))
+                logger.error(msg)
+                raise RuntimeError(msg)
+
+            parse_kwargs.update(kwargs)
 
         features = self._parse_features(features, **parse_kwargs)
 
