@@ -8,6 +8,7 @@ import os
 
 from phygnn.phygnn import PhysicsGuidedNeuralNetwork
 from phygnn.model_interfaces.base_model import ModelBase
+from phygnn.utilities.pre_processing import PreProcess
 
 logger = logging.getLogger(__name__)
 
@@ -16,8 +17,13 @@ class PhygnnModel(ModelBase):
     """
     Phygnn Model interface
     """
+
+    # Underlying model interface class. Used for loading models from disk
+    MODEL_CLASS = PhysicsGuidedNeuralNetwork
+
     def __init__(self, model, feature_names=None, label_names=None,
-                 norm_params=None, normalize=(True, False)):
+                 norm_params=None, normalize=(True, False),
+                 one_hot_categories=None):
         """
         Parameters
         ----------
@@ -37,10 +43,14 @@ class PhygnnModel(ModelBase):
             - False means don't normalize either
             - Tuple of flags (normalize_feature, normalize_label)
             by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
         """
         super().__init__(model, feature_names=feature_names,
                          label_names=label_names, norm_params=norm_params,
-                         normalize=normalize)
+                         normalize=normalize,
+                         one_hot_categories=one_hot_categories)
 
     @property
     def layers(self):
@@ -117,20 +127,19 @@ class PhygnnModel(ModelBase):
         Parameters
         ----------
         features : np.ndarray | pd.DataFrame
-            Feature data in a 2D array or DataFrame. If this is a DataFrame,
+            Feature data in a >=2D array or DataFrame. If this is a DataFrame,
             the index is ignored, the columns are used with self.feature_names,
             and the df is converted into a numpy array for batching and passing
-            to the training algorithm.
+            to the training algorithm. A 2D input should have the shape:
+            (n_observations, n_features). A 3D input should have the shape:
+            (n_observations, n_timesteps, n_features). 4D inputs have not been
+            tested and should be used with caution.
         labels : np.ndarray | pd.DataFrame
-            Known output data in a 2D array or DataFrame. If this is a
-            DataFrame, the index is ignored, the columns are used with
-            self.output_names, and the df is converted into a numpy array for
-            batching and passing to the training algorithm.
+            Known output data in a 2D array or DataFrame.
+            Same dimension rules as features.
         p : np.ndarray | pd.DataFrame
             Supplemental feature data for the physics loss function in 2D array
-            or DataFrame. If this is a DataFrame, the index and column labels
-            are ignored and the df is converted into a numpy array for batching
-            and passing to the training algorithm and physical loss function.
+            or DataFrame. Same dimension rules as features.
         n_batch : int
             Number of times to update the NN weights per epoch (number of
             mini-batches). The training data will be split into this many
@@ -140,15 +149,15 @@ class PhygnnModel(ModelBase):
             Number of times to iterate on the training data.
         shuffle : bool
             Flag to randomly subset the validation data and batch selection
-            from features and labels.
+            from features, labels, and p.
         validation_split : float
+            Fraction of features and labels to use for validation.
+        p_kwargs : None | dict
+            Optional kwargs for the physical loss function self._p_fun.
         run_preflight : bool
             Flag to run preflight checks.
         return_diagnostics : bool
             Flag to return training diagnostics dictionary.
-            Fraction of features and labels to use for validation.
-        p_kwargs : None | dict
-            Optional kwargs for the physical loss function self._p_fun.
         parse_kwargs : dict
             kwargs for cls._parse_features
         norm_labels : bool, optional
@@ -185,6 +194,8 @@ class PhygnnModel(ModelBase):
         path : str
             Save phygnn model
         """
+
+        path = os.path.abspath(path)
         if path.endswith(('.json', '.pkl')):
             dir_path = os.path.dirname(path)
             if path.endswith('.pkl'):
@@ -200,7 +211,8 @@ class PhygnnModel(ModelBase):
                         'label_names': self.label_names,
                         'norm_params': self.normalization_parameters,
                         'normalize': (self.normalize_features,
-                                      self.normalize_labels)}
+                                      self.normalize_labels),
+                        'one_hot_categories': self.one_hot_categories}
 
         model_params = self.dict_json_convert(model_params)
         with open(path, 'w') as f:
@@ -209,13 +221,38 @@ class PhygnnModel(ModelBase):
         path = path.replace('.json', '.pkl')
         self.model.save(path)
 
+    def set_loss_weights(self, loss_weights):
+        """Set new loss weights
+
+        Parameters
+        ----------
+        loss_weights : tuple
+            Loss weights for the neural network y_true vs y_predicted
+            and for the p_fun loss, respectively. For example,
+            loss_weights=(0.0, 1.0) would simplify the phygnn loss function
+            to just the p_fun output.
+        """
+        self.model._loss_weights = loss_weights
+
     @classmethod
     def build(cls, p_fun, feature_names, label_names,
-              normalize=(True, False), loss_weights=(0.5, 0.5),
-              hidden_layers=None, metric='mae', initializer=None,
-              optimizer=None, learning_rate=0.01, history=None,
-              kernel_reg_rate=0.0, kernel_reg_power=1,
-              bias_reg_rate=0.0, bias_reg_power=1):
+              normalize=(True, False),
+              one_hot_categories=None,
+              loss_weights=(0.5, 0.5),
+              hidden_layers=None,
+              input_layer=None,
+              output_layer=None,
+              layers_obj=None,
+              metric='mae',
+              initializer=None,
+              optimizer=None,
+              learning_rate=0.01,
+              history=None,
+              kernel_reg_rate=0.0,
+              kernel_reg_power=1,
+              bias_reg_rate=0.0,
+              bias_reg_power=1,
+              name=None):
         """
         Build phygnn model from given features, layers and kwargs
 
@@ -223,7 +260,7 @@ class PhygnnModel(ModelBase):
         ----------
         p_fun : function
             Physics function to guide the neural network loss function.
-            This fun must take (phygnn, y_predicted, y_true, p, **p_kwargs)
+            This fun must take (phygnn, y_true, y_predicted, p, **p_kwargs)
             as arguments with datatypes (PhysicsGuidedNeuralNetwork, tf.Tensor,
             np.ndarray, np.ndarray). The function must return a tf.Tensor
             object with a single numeric loss value (output.ndim == 0).
@@ -238,8 +275,11 @@ class PhygnnModel(ModelBase):
             - False means don't normalize either
             - Tuple of flags (normalize_feature, normalize_label)
             by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
         loss_weights : tuple, optional
-            Loss weights for the neural network y_predicted vs. y_true
+            Loss weights for the neural network y_true vs y_predicted
             and for the p_fun loss, respectively. For example,
             loss_weights=(0.0, 1.0) would simplify the phygnn loss function
             to just the p_fun output.
@@ -248,22 +288,40 @@ class PhygnnModel(ModelBase):
             layer in the NN. Dense linear layers can be input with their
             activations or separately for more explicit control over the layer
             ordering. For example, this is a valid input for hidden_layers that
-            will yield 7 hidden layers (9 layers total):
+            will yield 8 hidden layers (10 layers including input+output):
                 [{'units': 64, 'activation': 'relu', 'dropout': 0.01},
                  {'units': 64},
                  {'batch_normalization': {'axis': -1}},
                  {'activation': 'relu'},
-                 {'dropout': 0.01}]
+                 {'dropout': 0.01},
+                 {'class': 'Flatten'},
+                 ]
+        input_layer : None | dict
+            Input layer. specification. Can be a dictionary similar to
+            hidden_layers specifying a dense / conv / lstm layer.  Will
+            default to a keras InputLayer with input shape = n_features.
+        output_layer : None | list | dict
+            Output layer specification. Can be a list/dict similar to
+            hidden_layers input specifying a dense layer with activation.
+            For example, for a classfication problem with a single output,
+            output_layer should be [{'units': 1}, {'activation': 'sigmoid'}].
+            This defaults to a single dense layer with no activation
+            (best for regression problems).
+        layers_obj : None | phygnn.utilities.tf_layers.Layers
+            Optional initialized Layers object to set as the model layers
+            including pre-set weights. This option will override the
+            hidden_layers, input_layer, and output_layer arguments.
         metric : str, optional
             Loss metric option for the NN loss function (not the physical
             loss function). Must be a valid key in phygnn.loss_metrics.METRICS
         initializer : tensorflow.keras.initializers, optional
             Instantiated initializer object. None defaults to GlorotUniform
-        optimizer : tensorflow.keras.optimizers, optional
-            Instantiated neural network optimization object.
-            None defaults to Adam.
+        optimizer : tensorflow.keras.optimizers | dict | None
+            Instantiated tf.keras.optimizers object or a dict optimizer config
+            from tf.keras.optimizers.get_config(). None defaults to Adam.
         learning_rate : float, optional
-            Optimizer learning rate.
+            Optimizer learning rate. Not used if optimizer input arg is a
+            pre-initialized object or if optimizer input arg is a config dict.
         history : None | pd.DataFrame, optional
             Learning history if continuing a training session.
         kernel_reg_rate : float, optional
@@ -286,6 +344,8 @@ class PhygnnModel(ModelBase):
             Bias regularization power. bias_reg_power=1 is L1
             regularization (lasso regression), and bias_reg_power=2 is L2
             regularization (ridge regression).
+        name : None | str
+            Optional model name for debugging.
 
         Returns
         -------
@@ -295,11 +355,21 @@ class PhygnnModel(ModelBase):
         if isinstance(label_names, str):
             label_names = [label_names]
 
+        if one_hot_categories is not None:
+            check_names = feature_names + label_names
+            PreProcess.check_one_hot_categories(one_hot_categories,
+                                                feature_names=check_names)
+            feature_names = cls.make_one_hot_feature_names(feature_names,
+                                                           one_hot_categories)
+
         model = PhysicsGuidedNeuralNetwork(p_fun,
                                            loss_weights=loss_weights,
                                            n_features=len(feature_names),
                                            n_labels=len(label_names),
                                            hidden_layers=hidden_layers,
+                                           input_layer=input_layer,
+                                           output_layer=output_layer,
+                                           layers_obj=layers_obj,
                                            metric=metric,
                                            initializer=initializer,
                                            optimizer=optimizer,
@@ -310,23 +380,43 @@ class PhygnnModel(ModelBase):
                                            bias_reg_rate=bias_reg_rate,
                                            bias_reg_power=bias_reg_power,
                                            feature_names=feature_names,
-                                           output_names=label_names)
+                                           output_names=label_names,
+                                           name=name)
 
         model = cls(model, feature_names=feature_names,
-                    label_names=label_names, normalize=normalize)
+                    label_names=label_names, normalize=normalize,
+                    one_hot_categories=one_hot_categories)
 
         return model
 
     @classmethod
-    def build_trained(cls, p_fun, features, labels, p, normalize=(True, False),
-                      loss_weights=(0.5, 0.5), hidden_layers=None,
-                      metric='mae', initializer=None, optimizer=None,
-                      learning_rate=0.01, history=None,
-                      kernel_reg_rate=0.0, kernel_reg_power=1,
-                      bias_reg_rate=0.0, bias_reg_power=1, n_batch=16,
-                      n_epoch=10, shuffle=True, validation_split=0.2,
-                      run_preflight=True, return_diagnostics=False,
-                      p_kwargs=None, parse_kwargs=None, save_path=None):
+    def build_trained(cls, p_fun, features, labels, p,
+                      normalize=(True, False),
+                      one_hot_categories=None,
+                      loss_weights=(0.5, 0.5),
+                      hidden_layers=None,
+                      input_layer=None,
+                      output_layer=None,
+                      layers_obj=None,
+                      metric='mae',
+                      initializer=None,
+                      optimizer=None,
+                      learning_rate=0.01,
+                      history=None,
+                      kernel_reg_rate=0.0,
+                      kernel_reg_power=1,
+                      bias_reg_rate=0.0,
+                      bias_reg_power=1,
+                      n_batch=16,
+                      n_epoch=10,
+                      shuffle=True,
+                      validation_split=0.2,
+                      run_preflight=True,
+                      return_diagnostics=False,
+                      p_kwargs=None,
+                      parse_kwargs=None,
+                      save_path=None,
+                      name=None):
         """
         Build phygnn model from given features, layers and
         kwargs and then train with given labels and kwargs
@@ -335,25 +425,24 @@ class PhygnnModel(ModelBase):
         ----------
         p_fun : function
             Physics function to guide the neural network loss function.
-            This fun must take (phygnn, y_predicted, y_true, p, **p_kwargs)
+            This fun must take (phygnn, y_true, y_predicted, p, **p_kwargs)
             as arguments with datatypes (PhysicsGuidedNeuralNetwork, tf.Tensor,
             np.ndarray, np.ndarray). The function must return a tf.Tensor
             object with a single numeric loss value (output.ndim == 0).
         features : np.ndarray | pd.DataFrame
-            Feature data in a 2D array or DataFrame. If this is a DataFrame,
+            Feature data in a >=2D array or DataFrame. If this is a DataFrame,
             the index is ignored, the columns are used with self.feature_names,
             and the df is converted into a numpy array for batching and passing
-            to the training algorithm.
+            to the training algorithm. A 2D input should have the shape:
+            (n_observations, n_features). A 3D input should have the shape:
+            (n_observations, n_timesteps, n_features). 4D inputs have not been
+            tested and should be used with caution.
         labels : np.ndarray | pd.DataFrame
-            Known output data in a 2D array or DataFrame. If this is a
-            DataFrame, the index is ignored, the columns are used with
-            self.output_names, and the df is converted into a numpy array for
-            batching and passing to the training algorithm.
+            Known output data in a 2D array or DataFrame.
+            Same dimension rules as features.
         p : np.ndarray | pd.DataFrame
             Supplemental feature data for the physics loss function in 2D array
-            or DataFrame. If this is a DataFrame, the index and column labels
-            are ignored and the df is converted into a numpy array for batching
-            and passing to the training algorithm and physical loss function.
+            or DataFrame. Same dimension rules as features.
         normalize : bool | tuple, optional
             Boolean flag(s) as to whether features and labels should be
             normalized. Possible values:
@@ -361,8 +450,11 @@ class PhygnnModel(ModelBase):
             - False means don't normalize either
             - Tuple of flags (normalize_feature, normalize_label)
             by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
         loss_weights : tuple, optional
-            Loss weights for the neural network y_predicted vs. y_true
+            Loss weights for the neural network y_true vs y_predicted
             and for the p_fun loss, respectively. For example,
             loss_weights=(0.0, 1.0) would simplify the phygnn loss function
             to just the p_fun output.
@@ -371,22 +463,40 @@ class PhygnnModel(ModelBase):
             layer in the NN. Dense linear layers can be input with their
             activations or separately for more explicit control over the layer
             ordering. For example, this is a valid input for hidden_layers that
-            will yield 7 hidden layers (9 layers total):
+            will yield 8 hidden layers (10 layers including input+output):
                 [{'units': 64, 'activation': 'relu', 'dropout': 0.01},
                  {'units': 64},
                  {'batch_normalization': {'axis': -1}},
                  {'activation': 'relu'},
-                 {'dropout': 0.01}]
+                 {'dropout': 0.01},
+                 {'class': 'Flatten'},
+                 ]
+        input_layer : None | dict
+            Input layer. specification. Can be a dictionary similar to
+            hidden_layers specifying a dense / conv / lstm layer.  Will
+            default to a keras InputLayer with input shape = n_features.
+        output_layer : None | list | dict
+            Output layer specification. Can be a list/dict similar to
+            hidden_layers input specifying a dense layer with activation.
+            For example, for a classfication problem with a single output,
+            output_layer should be [{'units': 1}, {'activation': 'sigmoid'}].
+            This defaults to a single dense layer with no activation
+            (best for regression problems).
+        layers_obj : None | phygnn.utilities.tf_layers.Layers
+            Optional initialized Layers object to set as the model layers
+            including pre-set weights. This option will override the
+            hidden_layers, input_layer, and output_layer arguments.
         metric : str, optional
             Loss metric option for the NN loss function (not the physical
             loss function). Must be a valid key in phygnn.loss_metrics.METRICS
         initializer : tensorflow.keras.initializers, optional
             Instantiated initializer object. None defaults to GlorotUniform
-        optimizer : tensorflow.keras.optimizers, optional
-            Instantiated neural network optimization object.
-            None defaults to Adam.
+        optimizer : tensorflow.keras.optimizers | dict | None
+            Instantiated tf.keras.optimizers object or a dict optimizer config
+            from tf.keras.optimizers.get_config(). None defaults to Adam.
         learning_rate : float, optional
-            Optimizer learning rate.
+            Optimizer learning rate. Not used if optimizer input arg is a
+            pre-initialized object or if optimizer input arg is a config dict.
         history : None | pd.DataFrame, optional
             Learning history if continuing a training session.
         kernel_reg_rate : float, optional
@@ -435,6 +545,8 @@ class PhygnnModel(ModelBase):
             Directory path to save model to. The tensorflow model will be
             saved to the directory while the framework parameters will be
             saved in json, by default None
+        name : None | str
+            Optional model name for debugging.
 
         Returns
         -------
@@ -448,8 +560,12 @@ class PhygnnModel(ModelBase):
 
         model = cls.build(p_fun, feature_names, label_names,
                           normalize=normalize,
+                          one_hot_categories=one_hot_categories,
                           loss_weights=loss_weights,
                           hidden_layers=hidden_layers,
+                          input_layer=input_layer,
+                          output_layer=output_layer,
+                          layers_obj=layers_obj,
                           metric=metric,
                           initializer=initializer,
                           optimizer=optimizer,
@@ -458,7 +574,8 @@ class PhygnnModel(ModelBase):
                           kernel_reg_rate=kernel_reg_rate,
                           kernel_reg_power=kernel_reg_power,
                           bias_reg_rate=bias_reg_rate,
-                          bias_reg_power=bias_reg_power)
+                          bias_reg_power=bias_reg_power,
+                          name=name)
 
         diagnostics = model.train_model(features, labels, p,
                                         n_batch=n_batch,
@@ -505,7 +622,7 @@ class PhygnnModel(ModelBase):
             logger.error(e)
             raise IOError(e)
 
-        loaded = PhysicsGuidedNeuralNetwork.load(pkl_path)
+        loaded = cls.MODEL_CLASS.load(pkl_path)
 
         json_path = path.replace('.pkl', '.json')
         if not os.path.exists(json_path):
