@@ -13,7 +13,9 @@ from tensorflow.keras.optimizers import Adam
 from warnings import warn
 
 from phygnn.model_interfaces.base_model import ModelBase
+from phygnn.utilities import TF2
 from phygnn.utilities.tf_layers import Layers
+from phygnn.utilities.pre_processing import PreProcess
 
 logger = logging.getLogger(__name__)
 
@@ -23,7 +25,8 @@ class TfModel(ModelBase):
     TensorFlow Keras Model interface
     """
     def __init__(self, model, feature_names=None, label_names=None,
-                 norm_params=None, normalize=(True, False)):
+                 norm_params=None, normalize=(True, False),
+                 one_hot_categories=None):
         """
         Parameters
         ----------
@@ -43,10 +46,14 @@ class TfModel(ModelBase):
             - False means don't normalize either
             - Tuple of flags (normalize_feature, normalize_label)
             by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
         """
         super().__init__(model, feature_names=feature_names,
                          label_names=label_names, norm_params=norm_params,
-                         normalize=normalize)
+                         normalize=normalize,
+                         one_hot_categories=one_hot_categories)
 
         self._history = None
 
@@ -319,8 +326,9 @@ class TfModel(ModelBase):
 
         return model
 
-    def train_model(self, features, labels, epochs=100, validation_split=0.2,
-                    early_stop=True, parse_kwargs=None, fit_kwargs=None):
+    def train_model(self, features, labels, epochs=100, shuffle=True,
+                    validation_split=0.2, early_stop=True, parse_kwargs=None,
+                    fit_kwargs=None):
         """
         Train the model with the provided features and label
 
@@ -334,6 +342,9 @@ class TfModel(ModelBase):
             Flag to normalize label, by default True
         epochs : int, optional
             Number of epochs to train the model, by default 100
+        shuffle : bool
+            Flag to randomly subset the validation data and batch selection
+            from features and labels.
         validation_split : float, optional
             Fraction of the training data to be used as validation data,
             by default 0.2
@@ -369,6 +380,12 @@ class TfModel(ModelBase):
 
             fit_kwargs['callbacks'] = callbacks
 
+        if shuffle:
+            L = len(features)
+            i = np.random.choice(L, size=L, replace=False)
+            features = features[i]
+            labels = labels[i]
+
         if validation_split > 0:
             split = int(len(features) * validation_split)
             validate_features = features[-split:]
@@ -404,167 +421,23 @@ class TfModel(ModelBase):
         if not os.path.exists(path):
             os.makedirs(path)
 
-        tf.saved_model.save(self.model, path)
+        if TF2:
+            self.model.save(path)
+        if not TF2:
+            tf.saved_model.save(self.model, path)
 
         model_params = {'feature_names': self.feature_names,
                         'label_names': self.label_names,
                         'norm_params': self.normalization_parameters,
                         'normalize': (self.normalize_features,
-                                      self.normalize_labels)}
+                                      self.normalize_labels),
+                        'one_hot_categories': self.one_hot_categories}
 
-        json_path = path.rstrip('/') + '.json'
+        json_path = path + 'model.json'
         model_params = self.dict_json_convert(model_params)
         with open(json_path, 'w') as f:
             json.dump(model_params, f, indent=2, sort_keys=True)
 
-    @classmethod
-    def build(cls, feature_names, label_names, normalize=(True, False),
-              hidden_layers=None, learning_rate=0.001,
-              loss="mean_squared_error", metrics=('mae', 'mse'),
-              optimizer_class=Adam, **kwargs):
-        """
-        Build tensorflow sequential model from given features, layers and
-        kwargs
-
-        Parameters
-        ----------
-        feature_names : list
-            Ordered list of feature names.
-        label_names : list
-            Ordered list of label (output) names.
-        normalize : bool | tuple, optional
-            Boolean flag(s) as to whether features and labels should be
-            normalized. Possible values:
-            - True means normalize both
-            - False means don't normalize either
-            - Tuple of flags (normalize_feature, normalize_label)
-            by default True
-        hidden_layers : list, optional
-            List of tensorflow layers.Dense kwargs (dictionaries)
-            if None use a single linear layer, by default None
-        learning_rate : float, optional
-            tensorflow optimizer learning rate, by default 0.001
-        loss : str, optional
-            name of objective function, by default "mean_squared_error"
-        metrics : list, optional
-            List of metrics to be evaluated by the model during training and
-            testing, by default ('mae', 'mse')
-        optimizer_class : tf.keras.optimizers, optional
-            Optional explicit request of optimizer. This should be a class
-            that will be instantated in the TfModel._compile_model() method
-            The default is the Adam optimizer
-        kwargs : dict
-            kwargs for tensorflow.keras.models.compile
-
-        Returns
-        -------
-        model : TfModel
-            Initialized TfModel obj
-        """
-        if isinstance(label_names, str):
-            label_names = [label_names]
-
-        model = cls.compile_model(len(feature_names),
-                                  n_labels=len(label_names),
-                                  hidden_layers=hidden_layers,
-                                  learning_rate=learning_rate, loss=loss,
-                                  metrics=metrics,
-                                  optimizer_class=optimizer_class,
-                                  **kwargs)
-
-        model = cls(model, feature_names=feature_names,
-                    label_names=label_names, normalize=normalize)
-
-        return model
-
-    @classmethod
-    def build_trained(cls, features, labels, normalize=(True, False),
-                      hidden_layers=None, learning_rate=0.001,
-                      loss="mean_squared_error", metrics=('mae', 'mse'),
-                      optimizer_class=Adam, epochs=100, validation_split=0.2,
-                      early_stop=True, save_path=None, compile_kwargs=None,
-                      parse_kwargs=None, fit_kwargs=None):
-        """
-        Build tensorflow sequential model from given features, layers and
-        kwargs and then train with given label and kwargs
-
-        Parameters
-        ----------
-        features : dict | pandas.DataFrame
-            Model features
-        labels : dict | pandas.DataFrame
-            label to train on
-        normalize : bool | tuple, optional
-            Boolean flag(s) as to whether features and labels should be
-            normalized. Possible values:
-            - True means normalize both
-            - False means don't normalize either
-            - Tuple of flags (normalize_feature, normalize_label)
-            by default True
-        hidden_layers : list, optional
-            List of tensorflow layers.Dense kwargs (dictionaries)
-            if None use a single linear layer, by default None
-        learning_rate : float, optional
-            tensorflow optimizer learning rate, by default 0.001
-        loss : str, optional
-            name of objective function, by default "mean_squared_error"
-        metrics : list, optional
-            List of metrics to be evaluated by the model during training and
-            testing, by default ('mae', 'mse')
-        optimizer_class : tf.keras.optimizers, optional
-            Optional explicit request of optimizer. This should be a class
-            that will be instantated in the TfModel._compile_model() method
-            The default is the Adam optimizer
-        epochs : int, optional
-            Number of epochs to train the model, by default 100
-        validation_split : float, optional
-            Fraction of the training data to be used as validation data,
-            by default 0.2
-        early_stop : bool
-            Flag to stop training when it stops improving
-        save_path : str
-            Directory path to save model to. The tensorflow model will be
-            saved to the directory while the framework parameters will be
-            saved in json.
-        compile_kwargs : dict
-            kwargs for tensorflow.keras.models.compile
-        parse_kwargs : dict
-            kwargs for cls._parse_features
-        fit_kwargs : dict
-            kwargs for tensorflow.keras.models.fit
-
-        Returns
-        -------
-        model : TfModel
-            Initialized and trained TfModel obj
-        """
-        if compile_kwargs is None:
-            compile_kwargs = {}
-
-        _, feature_names = cls._parse_data(features)
-        _, label_names = cls._parse_data(labels)
-
-        model = cls.build(feature_names, label_names,
-                          normalize=normalize,
-                          hidden_layers=hidden_layers,
-                          learning_rate=learning_rate,
-                          loss=loss,
-                          metrics=metrics,
-                          optimizer_class=optimizer_class,
-                          **compile_kwargs)
-
-        model.train_model(features, labels,
-                          epochs=epochs,
-                          validation_split=validation_split,
-                          early_stop=early_stop,
-                          parse_kwargs=parse_kwargs,
-                          fit_kwargs=fit_kwargs)
-
-        if save_path is not None:
-            model.save_model(save_path)
-
-        return model
-
     @classmethod
     def load(cls, path):
         """
@@ -596,10 +469,178 @@ class TfModel(ModelBase):
 
         loaded = tf.keras.models.load_model(path)
 
-        json_path = path.rstrip('/') + '.json'
+        json_path = path + 'model.json'
         with open(json_path, 'r') as f:
             model_params = json.load(f)
 
         model = cls(loaded, **model_params)
 
         return model
+
+    @classmethod
+    def build(cls, feature_names, label_names, normalize=(True, False),
+              one_hot_categories=None, hidden_layers=None, learning_rate=0.001,
+              loss="mean_squared_error", metrics=('mae', 'mse'),
+              optimizer_class=Adam, **kwargs):
+        """
+        Build tensorflow sequential model from given features, layers and
+        kwargs
+
+        Parameters
+        ----------
+        feature_names : list
+            Ordered list of feature names.
+        label_names : list
+            Ordered list of label (output) names.
+        normalize : bool | tuple, optional
+            Boolean flag(s) as to whether features and labels should be
+            normalized. Possible values:
+            - True means normalize both
+            - False means don't normalize either
+            - Tuple of flags (normalize_feature, normalize_label)
+            by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
+        hidden_layers : list, optional
+            List of tensorflow layers.Dense kwargs (dictionaries)
+            if None use a single linear layer, by default None
+        learning_rate : float, optional
+            tensorflow optimizer learning rate, by default 0.001
+        loss : str, optional
+            name of objective function, by default "mean_squared_error"
+        metrics : list, optional
+            List of metrics to be evaluated by the model during training and
+            testing, by default ('mae', 'mse')
+        optimizer_class : tf.keras.optimizers, optional
+            Optional explicit request of optimizer. This should be a class
+            that will be instantated in the TfModel._compile_model() method
+            The default is the Adam optimizer
+        kwargs : dict
+            kwargs for tensorflow.keras.models.compile
+
+        Returns
+        -------
+        model : TfModel
+            Initialized TfModel obj
+        """
+        if isinstance(label_names, str):
+            label_names = [label_names]
+
+        if one_hot_categories is not None:
+            check_names = feature_names + label_names
+            PreProcess.check_one_hot_categories(one_hot_categories,
+                                                feature_names=check_names)
+            feature_names = cls.make_one_hot_feature_names(feature_names,
+                                                           one_hot_categories)
+
+        model = cls.compile_model(len(feature_names),
+                                  n_labels=len(label_names),
+                                  hidden_layers=hidden_layers,
+                                  learning_rate=learning_rate, loss=loss,
+                                  metrics=metrics,
+                                  optimizer_class=optimizer_class,
+                                  **kwargs)
+
+        model = cls(model, feature_names=feature_names,
+                    label_names=label_names, normalize=normalize,
+                    one_hot_categories=one_hot_categories)
+
+        return model
+
+    @classmethod
+    def build_trained(cls, features, labels, normalize=(True, False),
+                      one_hot_categories=None, hidden_layers=None,
+                      learning_rate=0.001, loss="mean_squared_error",
+                      metrics=('mae', 'mse'), optimizer_class=Adam, epochs=100,
+                      shuffle=True, validation_split=0.2, early_stop=True,
+                      save_path=None, compile_kwargs=None, parse_kwargs=None,
+                      fit_kwargs=None):
+        """
+        Build tensorflow sequential model from given features, layers and
+        kwargs and then train with given label and kwargs
+
+        Parameters
+        ----------
+        features : dict | pandas.DataFrame
+            Model features
+        labels : dict | pandas.DataFrame
+            label to train on
+        normalize : bool | tuple, optional
+            Boolean flag(s) as to whether features and labels should be
+            normalized. Possible values:
+            - True means normalize both
+            - False means don't normalize either
+            - Tuple of flags (normalize_feature, normalize_label)
+            by default True
+        one_hot_categories : dict, optional
+            Features to one-hot encode using given categories, if None do
+            not run one-hot encoding, by default None
+        hidden_layers : list, optional
+            List of tensorflow layers.Dense kwargs (dictionaries)
+            if None use a single linear layer, by default None
+        learning_rate : float, optional
+            tensorflow optimizer learning rate, by default 0.001
+        loss : str, optional
+            name of objective function, by default "mean_squared_error"
+        metrics : list, optional
+            List of metrics to be evaluated by the model during training and
+            testing, by default ('mae', 'mse')
+        optimizer_class : tf.keras.optimizers, optional
+            Optional explicit request of optimizer. This should be a class
+            that will be instantated in the TfModel._compile_model() method
+            The default is the Adam optimizer
+        epochs : int, optional
+            Number of epochs to train the model, by default 100
+        shuffle : bool
+            Flag to randomly subset the validation data and batch selection
+            from features and labels.
+        validation_split : float, optional
+            Fraction of the training data to be used as validation data,
+            by default 0.2
+        early_stop : bool
+            Flag to stop training when it stops improving
+        save_path : str
+            Directory path to save model to. The tensorflow model will be
+            saved to the directory while the framework parameters will be
+            saved in json.
+        compile_kwargs : dict
+            kwargs for tensorflow.keras.models.compile
+        parse_kwargs : dict
+            kwargs for cls._parse_features
+        fit_kwargs : dict
+            kwargs for tensorflow.keras.models.fit
+
+        Returns
+        -------
+        model : TfModel
+            Initialized and trained TfModel obj
+        """
+        if compile_kwargs is None:
+            compile_kwargs = {}
+
+        _, feature_names = cls._parse_data(features)
+        _, label_names = cls._parse_data(labels)
+
+        model = cls.build(feature_names, label_names,
+                          normalize=normalize,
+                          one_hot_categories=one_hot_categories,
+                          hidden_layers=hidden_layers,
+                          learning_rate=learning_rate,
+                          loss=loss,
+                          metrics=metrics,
+                          optimizer_class=optimizer_class,
+                          **compile_kwargs)
+
+        model.train_model(features, labels,
+                          epochs=epochs,
+                          shuffle=shuffle,
+                          validation_split=validation_split,
+                          early_stop=early_stop,
+                          parse_kwargs=parse_kwargs,
+                          fit_kwargs=fit_kwargs)
+
+        if save_path is not None:
+            model.save_model(save_path)
+
+        return model
